//! Custom implementation of the Vector datastructure
//!
//! Many thanks to the rustonomicon, chapter 9:
//! https://doc.rust-lang.org/nomicon/vec/vec.html

use std::{
    alloc::{self, Layout},
    marker::PhantomData,
    mem,
    ops::{Deref, DerefMut, Index},
    ptr::{self, NonNull},
};

use crate::raw_vec::RawVec;

#[derive(Clone, Debug)]
pub struct Vec<T> {
    used: usize,
    marker: PhantomData<T>,
    buf: RawVec<T>,
}

impl<T> Default for Vec<T> {
    fn default() -> Self {
        Self::new()
    }
}

impl<T> Vec<T> {
    pub fn new() -> Self {
        if mem::size_of::<T>() == 0 {
            panic!("We're not ready to handle ZSTs");
        }
        Vec {
            used: 0,
            marker: PhantomData::default(),
            buf: RawVec::new(),
        }
    }

    pub fn pop(&mut self) -> Option<T> {
        if self.used == 0 {
            None
        } else {
            self.used -= 1;
            unsafe { Some(ptr::read(self.buf.ptr.as_ptr().add(self.used))) }
        }
    }

    #[must_use]
    pub fn get(&self, index: usize) -> &T {
        todo!()
    }

    #[must_use]
    pub fn get_mut(&self, index: usize) -> &mut T {
        todo!()
    }

    pub fn push(&mut self, value: T) {
        if self.used == self.buf.capacity {
            self.buf.grow();
        }

        unsafe {
            ptr::write(self.buf.ptr.as_ptr().add(self.used), value);
        }

        self.used += 1;
    }

    #[must_use]
    pub fn len(&self) -> usize {
        self.used
    }

    #[must_use]
    pub fn capacity(&self) -> usize {
        self.buf.capacity
    }

    pub fn is_empty(&self) -> bool {
        self.len() == 0
    }

    pub fn insert(&mut self, index: usize, elem: T) {
        // Note: `<=` because it's valid to insert after everything
        // which would be equivalent to push.
        assert!(index <= self.used, "index out of bounds");
        if self.used == self.buf.capacity {
            self.buf.grow();
        }

        unsafe {
            // ptr::copy(src, dest, len): "copy from src to dest len elems"
            ptr::copy(
                self.buf.ptr.as_ptr().add(index),
                self.buf.ptr.as_ptr().add(index + 1),
                self.used - index,
            );
            ptr::write(self.buf.ptr.as_ptr().add(index), elem);
        }

        self.used += 1;
    }

    pub fn remove(&mut self, index: usize) -> T {
        // Note: `<` because it's *not* valid to remove after everything
        assert!(index < self.used, "index out of bounds");
        unsafe {
            self.used -= 1;
            let result = ptr::read(self.buf.ptr.as_ptr().add(index));
            ptr::copy(
                self.buf.ptr.as_ptr().add(index + 1),
                self.buf.ptr.as_ptr().add(index),
                self.used - index,
            );
            result
        }
    }
}

impl<T> Index<usize> for Vec<T> {
    type Output = T;

    fn index(&self, index: usize) -> &Self::Output {
        self.get(index)
    }
}

impl<T> Drop for Vec<T> {
    fn drop(&mut self) {
        if self.buf.capacity != 0 {
            while self.pop().is_some() {}
            let layout = Layout::array::<T>(self.buf.capacity).unwrap();
            unsafe {
                alloc::dealloc(self.buf.ptr.as_ptr() as *mut u8, layout);
            }
        }
    }
}

impl<T> Deref for Vec<T> {
    type Target = [T];
    fn deref(&self) -> &[T] {
        unsafe { std::slice::from_raw_parts(self.buf.ptr.as_ptr(), self.used) }
    }
}

impl<T> DerefMut for Vec<T> {
    fn deref_mut(&mut self) -> &mut [T] {
        unsafe { std::slice::from_raw_parts_mut(self.buf.ptr.as_ptr(), self.used) }
    }
}

unsafe impl<T: Send> Send for Vec<T> {}
unsafe impl<T: Sync> Sync for Vec<T> {}

#[cfg(test)]
mod test {
    use super::*;
    #[test]
    fn test_create() {
        let _v = Vec::<u64>::new();
    }

    #[test]
    fn test_pushpop_num() {
        let mut v = Vec::new();
        let vals = &[19, 9, 14, 255, 19191919, 13890, 21521, 1251, 6216, 1830];

        for val in vals {
            v.push(*val);
        }
        for val in vals.iter().rev() {
            assert_eq!(v.pop().unwrap(), *val);
        }
    }

    #[test]
    fn test_pushpop_str() {
        let mut v = Vec::new();
        let vals = &["AAAA", "ABBAB", "BBABBABBAJJJ"];

        for val in vals {
            v.push(*val);
        }
        for val in vals.iter().rev() {
            assert_eq!(v.pop().unwrap(), *val);
        }
    }
}
